@@ -248,6 +248,40 @@ class PowerPoint():
 
                     except IndexError:
                         self.log.debug(f'index error in update_table ({name}) while setting values')
+    
+    def set_table_bgcolor(self,table,colors,rows,cols,has_header):
+        for row in range(rows):
+            rgb = colors.iloc[row].split(",")
+            for col in range(cols):
+                try:
+                    if has_header:
+                        cell = table.cell(row+1,col)
+                    else:
+                        cell = table.cell(row,col)
+                    cell.fill.solid()
+                    cell.fill.fore_color.rgb = RGBColor(int(rgb[0]), int(rgb [1]), int(rgb[2]))
+                except IndexError:
+                    self.warning('index error in update_table while setting background color')
+    
+    def set_table_font_color(self,table,colors,rows,cols,has_header):
+        for row in range(rows):
+            rgb = colors.iloc[row].split(",")
+            for col in range(cols):
+                try:
+                    if has_header:
+                        cell = table.cell(row+1,col)
+                    else:
+                        cell = table.cell(row,col)
+
+                    paragraph = cell.text_frame.paragraphs[0]
+                    run = self._merge_runs(paragraph)
+                    run.font.color.rgb=RGBColor(int(rgb[0]), int(rgb [1]), int(rgb[2]))
+                except IndexError:
+                    self.warning('index error in update_table while setting background color')
+
+    def change_paragraph_color(self,paragraph,rgb):
+        run = self._merge_runs(paragraph)
+        run.font.color.rgb=RGBColor(int(rgb[0]), int(rgb [1]), int(rgb[2]))
 
     def isFloat(str):
         rslt = True
@@ -278,8 +312,8 @@ class PowerPoint():
                         for paragraph in shape.text_frame.paragraphs:
                             if paragraph.text == "{app_per_page}":
                                 new_slide = self.copy_slide(idx)
-                                self.replace_slide_text(new_slide,"{app_per_page}","")
-                                self.replace_slide_text(new_slide,"{app1_",f'{{app{cnt}_')
+                                self._replace_slide_text(new_slide,"{app_per_page}","")
+                                self._replace_slide_text(new_slide,"{app1_",f'{{app{cnt}_')
                                 self.replace_shape_name(new_slide,"app1_",f'app{cnt}_')
                             # if paragraph.text == "{multi_page}":
                             #     new_slide = self.copy_slide(idx)
@@ -292,7 +326,7 @@ class PowerPoint():
                 if shape.has_text_frame:
                     for paragraph in shape.text_frame.paragraphs:
                         if paragraph.text == "{app_per_page}":
-                            self.replace_slide_text(slide,"{app_per_page}","")
+                            self._replace_slide_text(slide,"{app_per_page}","")
 
     def copy_slide(self,index=-1,template=None):
         if index<0 and template is None:
@@ -333,3 +367,107 @@ class PowerPoint():
             except AttributeError as err:
                 self.logger.error(f'Attribute Error {err} while copying slide {index} part {key}')
         return dest
+
+    def copy_block(self, tag, prefix, count,slide=None):
+        search_start = f'{{{tag}}}'
+        search_end = f'{{end_{tag}}}'
+
+        block = []
+
+        slides = []
+        if slide is not None:
+            slides = [slide] 
+        else:
+            slides = self._prs.slides
+
+
+        found=False
+        for slide in slides:
+            for shape in slide.shapes:
+                if shape.has_text_frame:
+                    for paragraph in shape.text_frame.paragraphs:
+                        text=paragraph.text
+                        if not found and text.find(search_start)!=-1:
+                            #is the end in the same paragraph?
+                            if text.find(search_end)!=-1:
+                                run=self._merge_runs(paragraph)
+                                old_text = run.text
+                                text_prefix = text[:old_text.find(search_end)]
+                                text_suffix = text[old_text.find(search_end):]
+
+                                sub_text = ""
+                                for app_no in range (2,count+1):
+                                    temp = old_text[old_text.find(search_start)+len(search_start):old_text.find(search_end)]
+                                    for p in prefix:
+                                        temp = temp.replace(f'{p}1',f'{p}{app_no}')
+                                    sub_text = sub_text + ", " + temp
+                                
+                                new_text = text_prefix.replace(search_start,'') + sub_text + text_suffix.replace(search_end,'')
+                                #new_text = text_prefix + sub_text + text_suffix
+
+                                run.text = run.text.replace(old_text,new_text)
+                            else:
+                                found=True
+                        elif found and text.find(search_end)!=-1:
+                            found=False
+                            for app_no in range(2,count+1):
+                                self.paste_block(block, shape.text_frame,app_no)
+                            if paragraph.text==search_end:
+                                self.delete_paragraph(paragraph)
+                            block=[]
+                        if found:
+                            if paragraph.text==search_start:
+                                self.delete_paragraph(paragraph)
+                            else:
+                                block.append(paragraph)
+
+    def delete_paragraph(self,paragraph):
+        p = paragraph._p
+        parent_element = p.getparent()
+        parent_element.remove(p)
+
+    def paste_block(self,block, text_frame,app_no):
+        start_at = block[-1]
+
+        for b in block:
+            p = text_frame.add_paragraph()
+            p.alignment = b.alignment
+            p.line_spacing = b.line_spacing
+            p.level = b.level
+
+            for r in b.runs:    
+                run = p.add_run()
+                run.text = deepcopy(r.text)
+                font = run.font
+                font.name = r.font.name
+                font.size = r.font.size
+                font.bold = r.font.bold
+                font.italic = r.font.italic
+                if hasattr(r.font.color, 'rgb'):
+                    font.color.rgb = r.font.color.rgb
+                # else:
+                #     font.color.theme_color = r.font.color.theme_color 
+
+            run = self.merge_runs(p)
+            run.text = run.text.replace("{app1_",f'{{app{app_no}_')
+            run.text = run.text.replace("{end_app1_",f'{{end_app{app_no}_')
+
+    def merge_runs(self, paragraph):
+        cur_text=''
+        first=True
+        for run in paragraph.runs:
+            cur_text = cur_text + run.text
+            if first != True:
+                self.delete_run(run)
+            first=False
+        if len(paragraph.runs) > 0:   
+            run = paragraph.runs[0]
+            run.text = cur_text
+        else: 
+            run = paragraph.add_run()
+        return run
+
+    def replace_shape_name (self, slide, search_str, repl_str):
+        for shape in slide.shapes:
+            if shape.name.find(search_str) != -1: 
+                shape.name = shape.name.replace(search_str,repl_str)
